#include "chil_global.h"


/* Write a filename to dest that complies with the specified format,
   which may contain the following codes:
    %f   4-digit file (or tape) number 
    %r   3-digit raster number
    %s   2-digit scan number
    %d   date as YYYYMMDD
    %t   time as HHMMSS
    %N   system name
    %n   system name in lower case
    %C   3-letter scan-type code (RHI, PPI etc)
    %c   3-letter scan-type code in lower case
    %%   the character %
   Returns the number of letters in the filename (0 if an error occurred).
*/
int
CHILCreateFileName(char *dest, char *format, char *prefix,
		   uword file, uword raster, uword scan,
		   uword year, uword month, uword day,
		   uword hour, uword minute, uword second,
		   char *system, char *scantype)
{
  char *dest_save = dest;
  if (prefix) {
    int len;
    strncpy(dest, prefix, MAX_FILENAME_LENGTH);
    len = strlen(dest);
    dest += len;
    if (len > MAX_FILENAME_LENGTH-2) {
      *dest = '\0';
      /* Not enough room for filename */
      return 0;
    }
    else {
      *(dest++) = '/';
    }
  }
  while ((dest - dest_save) < MAX_FILENAME_LENGTH) {
    char ch = *(format++);
    if (ch == '\0') {
      *dest = '\0';
      return (dest - dest_save);
    }
    else if (ch < ' ') {
      /* Illegal character */
      *dest = '\0';
      return 0;
    }
    else if (ch == '%') {
      char *l;
      char code = *(format++);
      switch (code) {
      case 'f':
	snprintf(dest, 5, "%04hu", file);
	dest += 4;
	break;
      case 'r':
	snprintf(dest, 4, "%03hu", raster);
	dest += 3;
	break;
      case 's':
	snprintf(dest, 3, "%02hu", scan);
	dest += 2;
	break;
      case 'd':
	snprintf(dest, 9, "%04hu%02hu%02hu", CHILTrueYear(year), month, day);
	dest += 8;
	break;
      case 't':
	snprintf(dest, 7, "%02hu%02hu%02hu", hour, minute, second);
	dest += 6;
	break;
      case 'N':
	l = system;
	while (*l) {
	  *(dest++) = *(l++);
	}
	break;
      case 'n':
	l = system;
	while (*l) {
	  *(dest++) = tolower(*(l++));
	}
	break;
      case 'C':
	switch (*scantype) {
	case 'F': case 'f':
	  strcpy(dest, "FIX");
	  dest += 3;
	  break;
	case 'R': case 'r':
	  strcpy(dest, "RHI");
	  dest += 3;
	  break;
	case 'P': case 'p':
	  strcpy(dest, "PPI");
	  dest += 3;
	  break;
	case 'S': case 's':
	  strcpy(dest, "S-P");
	  dest += 3;
	  break;
	case 'C': case 'c':
	  strcpy(dest, "CAL");
	  dest += 3;
	  break;
	case 'M': case 'm':
	  strcpy(dest, "MAN");
	  dest += 3;
	  break;
	}
	break;
      case 'c':
	switch (*scantype) {
	case 'F': case 'f':
	  strcpy(dest, "fix");
	  dest += 3;
	  break;
	case 'R': case 'r':
	  strcpy(dest, "rhi");
	  dest += 3;
	  break;
	case 'P': case 'p':
	  strcpy(dest, "ppi");
	  dest += 3;
	  break;
	case 'S': case 's':
	  strcpy(dest, "s-p");
	  dest += 3;
	  break;
	case 'C': case 'c':
	  strcpy(dest, "cal");
	  dest += 3;
	  break;
	case 'M': case 'm':
	  strcpy(dest, "man");
	  dest += 3;
	  break;
	}
	break;
      case '%':
	*(dest++) = '%';
	break;
      default:
	*dest = '\n';
	/* Unrecognised code */
	return 0;
      }
      
    }
    else {
      *(dest++) = ch;
    }
  }
  *dest = '\n';
  return 0;
}

void
CHILCreateFileNameUsage()
{
  CHILPrintf("     -filename s  write filenames based on string s, containing any of\n"
	     "                     %%f   4-digit file (or tape) number\n"
	     "                     %%r   3-digit raster number\n"
	     "                     %%s   2-digit scan number\n"
	     "                     %%d   date as YYYYMMDD\n"
	     "                     %%t   time as HHMMSS\n"
	     "                     %%N   system name\n"
	     "                     %%n   system name in lower case\n"
	     "                     %%C   3-letter scan-type code (RHI, PPI etc)\n"
	     "                     %%c   3-letter scan-type code in lower case\n"
	     "                     %%%%   the character %%\n");
}
