/*
  loadceil.c

  loads ceilometer data from a single file into matlab - compile with mex

  [utc,range,beta,cloud1,cloud2,cloud3] = ...
     loadceil(ceilfilename,averays,avegates,tfactor);
*/

#include <stdio.h>
#include "mex.h"
#include "ceil.h"
#define PutValue(value,array,ray) *((double *) mxGetPr(array) + (ray))=value
#define PutCloudBase(value,array,ray) *((double *) mxGetPr(array) + (ray))=(value==0 ? nan : (double) value)

double nan;

void
PutRay(LCBackscatterProfile *profile,mxArray *array,int ray,
       int mustthreshold,int minbeta) {
  int ngates,i,j;
  double *data;

  ngates = ( mxGetM(array) < profile->ngates ) 
    ? mxGetM(array) : profile->ngates;
  data=mxGetPr(array);
  j=ray*mxGetM(array);

  if (mustthreshold) {
    for (i=0;i<ngates;i++) {
      if (profile->beta[i]<minbeta)
	data[i+j]=nan;
      else
	data[i+j]=profile->beta[i];
    }
  }
  else {
    for (i=0;i<ngates;i++)
      data[i+j]=profile->beta[i];
  }
}

void
PutRange(LCBackscatterProfile *profile,mxArray *array) {
  int ngates,i;
  double *data;

  ngates = ( mxGetN(array) < profile->ngates ) 
    ? mxGetN(array) : profile->ngates;
  data=mxGetPr(array);

  for (i=0;i<ngates;i++) {
    data[i]=profile->range[i];
  }
}

void
mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[]) {

  LCTime time; 
  LCCloudBases cloudbases; 
  LCBackscatterProfile profile_hr,profile; 

  int cloudgate; 
  float stdn; 
  int status;

  char *filename = NULL; int averays=1,avegates=1; float tfactor=4.0; 
  int ok=1;
  FILE *ceilfile;
  int autc=0,arange=1,abeta=2,ac1=3,ac2=4,ac3=5,adate=6;
  int ngates=0,nrays=740,ray=0;
  int buflen;
  double utc,date;
  int mustclean=1,mustthreshold=0;
  int minbeta=1;
  int noisegates=8;

  nan=mxGetNaN();
  /* process input arguments */
  if (nrhs < 1) {
    mexPrintf("USAGE:\n\n[utc,range,beta,c1,c2,c3,date]=...\n"
	      "   loadceil(filename,averays,avegates,tfactor,minbeta);\n\n");
    mexPrintf("All input arguments are optional except filename\n");
    mexPrintf("      averays: number of rays to average (default %d)\n",
	      averays);
    mexPrintf("      avegates: number of gates to average (default %d)\n",
	      avegates);
    mexPrintf("      tfactor: number of standard deviations above the noise\n            above which a signal at the top of the ray is accepted;\n"
	      "            negative values mean don't clean ray at all (default %f)\n",
	      tfactor);
    mexPrintf("      minbeta: backscatter below this value is set to NaN\n"
	      "            if not set the effectively -Inf; usual values are 1 or 2\n");
    mexErrMsgTxt("Not enough input arguments\n");
    ok=0;
  }
  else {
    if (mxIsChar(prhs[0])) {
      buflen = (mxGetM(prhs[0]) * mxGetN(prhs[0])) + 1;
      filename=mxCalloc(buflen, sizeof(char));
      mxGetString(prhs[0],filename,buflen);
    }
    else {
      mexErrMsgTxt("Filename must be a string\n");
      ok=0;
    }
    if (nrhs > 1) {
      averays=(int) *mxGetPr(prhs[1]);
      nrays /= averays;
      if (nrhs > 2) {
	avegates=(int) *mxGetPr(prhs[2]);
	if (nrhs > 3) {
	  tfactor=(float) *mxGetPr(prhs[3]);
	  if (tfactor <=0 ) mustclean=0;
	  if (nrhs > 4) {
	    minbeta=(int) *mxGetPr(prhs[4]);
	    mustthreshold=1;
	  }
	}
      }
    }
  }

  if (nlhs < 2) {
    abeta=0;
  }
  else if (nlhs == 2) {
    mexErrMsgTxt("Must have 1, 3, 4 or 5 output arguments\n");
    ok=0;
  }


  if (ok) {
    if ((ceilfile=fopen(filename,"r")) == NULL) {
      mexErrMsgTxt("File not found\n");
      ok=0;
    }
    else {
      if ((status=LCGetAveragedRay(ceilfile,&time,&cloudbases,&profile_hr,
				   averays)) > 0) {
	LCAverageRay(&profile_hr,&profile,avegates);
	if (mustclean) 
	  LCCustomCleanRay(&cloudgate,&stdn,&profile,noisegates,tfactor);
	ngates=profile.ngates;
	plhs[abeta]=mxCreateDoubleMatrix(ngates,nrays,mxREAL);
	PutRay(&profile,plhs[abeta],ray++,mustthreshold,minbeta);
	if (nlhs > 2) {
	  plhs[autc]=mxCreateDoubleMatrix(1,nrays,mxREAL);
	  utc= ((double) time.hour)+((double) time.minute)/60.0+
	    ((double) time.second)/3600.0;
	  PutValue(utc,plhs[autc],0);
	  plhs[arange]=mxCreateDoubleMatrix(1,ngates,mxREAL);
	  PutRange(&profile,plhs[arange]);
	  if (nlhs >3) {
	    plhs[ac1]=mxCreateDoubleMatrix(1,nrays,mxREAL);
	    PutCloudBase(cloudbases.height1,plhs[ac1],0);
	    if (nlhs >4) {
	      plhs[ac2]=mxCreateDoubleMatrix(1,nrays,mxREAL);
	      PutCloudBase(cloudbases.height2,plhs[ac2],0);
	      if (nlhs >5) {
		plhs[ac3]=mxCreateDoubleMatrix(1,nrays,mxREAL);
		PutCloudBase(cloudbases.height3,plhs[ac3],0);
		if (nlhs >6) {
		  plhs[adate]=mxCreateDoubleMatrix(1,nrays,mxREAL);
		  date=time.year*10000+time.month*100+time.day;
		  PutValue(date,plhs[adate],0);
		}
	      }
	    }
	  }
	}

	while (LCGetAveragedRay(ceilfile,&time,&cloudbases,&profile_hr,
				averays) > 0 && ray<nrays) {
	  LCAverageRay(&profile_hr,&profile,avegates);
	  if (mustclean) LCCustomCleanRay(&cloudgate,&stdn,&profile,
					  noisegates,tfactor);
	  if (nlhs > 2) {
	    utc= ((double) time.hour)+((double) time.minute)/60.0+
	      ((double) time.second)/3600.0;
	    PutValue(utc,plhs[autc],ray);
	    if (nlhs >3) {
	      PutCloudBase(cloudbases.height1,plhs[ac1],ray);
	      if (nlhs >4) {
		PutCloudBase(cloudbases.height2,plhs[ac2],ray);
		if (nlhs >5) {
		  PutCloudBase(cloudbases.height3,plhs[ac3],ray);
		  if (nlhs >6) {
		    date=time.year*10000+time.month*100+time.day;
		    PutValue(date,plhs[adate],ray);
		  }
		}
	      }
	    }
	  }
	  PutRay(&profile,plhs[abeta],ray++,mustthreshold,minbeta);
	}
	mxSetN(plhs[abeta],ray);
	if (nlhs > 2) {
	  mxSetN(plhs[autc],ray);
	  if (nlhs >3) {
	    mxSetN(plhs[ac1],ray);
	    if (nlhs >4) {
	      mxSetN(plhs[ac2],ray);
	      if (nlhs >5) {
		mxSetN(plhs[ac3],ray);
		if (nlhs >6) {
		  mxSetN(plhs[adate],ray);
		}
	      }
	    }
	  }
	}
      }
      else {
        plhs[abeta]=mxCreateDoubleMatrix(0,0,mxREAL);
        plhs[autc]=mxCreateDoubleMatrix(0,0,mxREAL);
        plhs[arange]=mxCreateDoubleMatrix(0,0,mxREAL);
	if (nlhs >3) {
	  plhs[ac1]=mxCreateDoubleMatrix(0,0,mxREAL);
	  if (nlhs >4) {
	    plhs[ac2]=mxCreateDoubleMatrix(0,0,mxREAL);
	    if (nlhs >5) {
	      plhs[ac3]=mxCreateDoubleMatrix(0,0,mxREAL);
	      if (nlhs >6) {
		plhs[adate]=mxCreateDoubleMatrix(0,0,mxREAL);
	      }
	    }
	  }
	}
      }
      fclose(ceilfile);
      mxFree(filename);
    }
  }
}

          

