#include <iostream>
#include <fstream>
#include <cstring>
#include <cstdio>
#include <cmath>
#include "rplot.h"

extern "C" {
#include <netcdf.h>
}

int ncstatus;
#define handle_error(a) if ((ncstatus = (a)) != NC_NOERR) { \
    cerr << "NetCDF error on line " << __LINE__ << ": " \
      << nc_strerror(ncstatus) << "\n"; \
    exit(-1); \
} 

#define MAX_PLOTS 16
#define NO_SIGNAL -999.0
#define NO_FLOAT_ATTRIBUTE -1.0e30
#define NO_FLOAT_ATTRIBUTE_TEST -0.9e30

#define DEFAULT_RAY_DURATION 0.01

//#define DEBUG cerr << "Reached line " << __LINE__ << "\n"
#define DEBUG /* */

enum FILE_FORMAT {
  FORMAT_POSTSCRIPT,
  FORMAT_PNG
};

static bool verbose = false, quiet = false;

static
char *
get_char_attribute(int ncid, int varid, char *name)
{
  size_t length;
  nc_type type;
  char *value;

  handle_error(nc_inq_atttype(ncid, varid, name, &type));
  if (type != NC_CHAR) {
    cerr << "Error: attribute " << name << " is not of type char\n";
    exit(-1);
  }
  handle_error(nc_inq_attlen(ncid, varid, name, &length));
  value = new char[length+1];
  if (!value) {
    cerr << "Out of memory\n";
    exit(2);
  }
  handle_error(nc_get_att_text(ncid, varid, name, value));
  value[length] = '\0';
  return value;
}

static
short int
get_short_attribute(int ncid, int varid, char *name)
{
  size_t length;
  nc_type type;
  short int value;

  handle_error(nc_inq_atttype(ncid, varid, name, &type));
  if (type != NC_SHORT) {
    cerr << "Error: attribute " << name << " is not of type short int\n";
    exit(-1);
  }
  handle_error(nc_inq_attlen(ncid, varid, name, &length));
  if (length > 1) {
    cerr << "Error: attribute " << name << " has more than 1 element\n";
  }
  handle_error(nc_get_att_short(ncid, varid, name, &value));
  return value;
}

static
float
get_float_attribute(int ncid, int varid, char *name)
{
  size_t length;
  float value;

  if (nc_inq_attlen(ncid, varid, name, &length) != NC_NOERR) {
    return NO_FLOAT_ATTRIBUTE;
  }
  if (length > 1) {
    cerr << "Error: " << name << " attribute has more than 1 element\n";
  }
  handle_error(nc_get_att_float(ncid, varid, name, &value));
  return value;
}

static
int
get_date(char *datestr, int day, int month, int year)
{
  static char *months[] = {"Jan","Feb","Mar","Apr","May","June",
			   "July","Aug","Sept","Oct","Nov","Dec"};
  if (month < 1 || month > 12) {
    return 1;
  }
  else {
    sprintf(datestr, "%d %s %d", day, months[month-1], year);
  }
  return 0;
}

static
void
Usage(char **argv)
{
  cerr << "Usage: " << argv[0] << " [options] file1.nc [file2.nc ...]\n"
    "Options:\n"
    "    -hourrange hr1 hr2  set hour axis range\n"
    "  -heightrange ht1 ht2  set height axis range\n"
    "        -range f1 f2    set colour scale range for all plotted parameters\n"
    "       -Zrange Z1 Z2    set colour scale range for Z (dBZ)\n"
    "       -vrange v1 v2    set colour scale range for velocity (m/s)\n"
    "   -widthrange w1 w2    set colour scale range for spectral width (m/s)\n"
    "      -svrange sv1 sv2  set colour scale range for velocity standard deviation (m/s)\n"
    "       -nezcal f        calibrate radar by forcing nez to be f dBZ\n"
    "      -betacal f        calibration figure for lidar beta (dB)\n"
    "     -duration hrs      plot only specified number of hours - if negative,\n"
    "                          plot most recent data (overridden by `-hourrange')\n"
    "  -rayduration f        assumed duration of each ray in seconds (default "
       << DEFAULT_RAY_DURATION*3600.0 << ")\n"
    "         -size <x>x<y>  set size of axes in pixels\n"
    "      -barsize <x>x<y>  set size of colourbar in pixels\n"
    "     -fontsize n        set pixel size of font\n"
    "       -system system   specify the name of the instrument\n"
    "        -field field    specify the full name of field being plotted\n"
    "        -units units    specify the units of the field\n"
    "           -id id       string identifier used in the output filename\n"
    "     -onefield          plot only the first field from each file\n"
    "      -oneplot          produce only one plot and send to standard output\n"
    "          -png          set image format to portable network graphics (default)\n"
    "     -ps, -eps          set image format to encapsulated postscript\n"
    "       -prefix dir      write the plots in specified directory\n"
    "      -palette file     read RGB colourmap from a file\n"
    "          -xfs server   specify the X font server\n" 
    "      -verbose          talk to me\n";
  exit(0);
}

int
main(int argc, char **argv)
{
  RFigure *figure[MAX_PLOTS];
  RAxes *axes[MAX_PLOTS];
  RPalette colormap;
  RFontLoader *font_loader;
  ofstream ofs[MAX_PLOTS];

  FILE_FORMAT format=FORMAT_PNG;
  char *xfs = "";
  int width = 650, height = 230;
  int bar_width = 16, bar_height = 160;
  int x_offset = 50, y_offset = 40;
  int fig_width, fig_height;

  int iarg, ifile, ivar;
  int max_vars, nvars;
  int ncid, rangeid, rangedimid, timedimid, timeid, varid[MAX_PLOTS], nezid = 0;
  size_t ngates, nrays, iray;
  float *range, *var, time, nexttime, nez, nez_override = 0.0;
  size_t start[2], count[2];
  int tmp;
  short int year, month, day;

  float minvalue[MAX_PLOTS], maxvalue[MAX_PLOTS], interval[MAX_PLOTS];
  float missing_value[MAX_PLOTS], vfold[MAX_PLOTS];
  float minval_override = 0.0, maxval_override = 0.0;
  float minZ = -50.0, maxZ = 10.0;
  float minv = -5.0, maxv = 5.0;
  /* float minwidth = 0.0, maxwidth = 0.8; // This is for a linear scale */
  float minwidth = -1.5, maxwidth = 0; /* This is for a logarithmic scale */
  /* float minsv = 0.0, maxsv = 2.0; // This is for a linear scale */
  float minsv = -3, maxsv = 0; /* This is for a logarithmic scale */
  float betacal = 0.0; /* In bels!!! */
  bool log_scale[MAX_PLOTS], using_nez[MAX_PLOTS];
  char *system=NULL, *system_override=NULL;
  char *long_name=NULL, *long_name_override=NULL;
  char *units=NULL, *units_override=NULL;
  char *suffix, *underscore1 = "", *underscore2 = "";
  char *system_code = "", *prefix = ".", *mapfile = NULL;
  int font_size = 12;
  float starthour = 0, endhour = 24, minheight = 0, maxheight = 12;
  float duration = 0, dhour, rayduration = DEFAULT_RAY_DURATION;
  bool no_bar = false, one_field = false, one_plot = false;
  bool use_duration = false, override_range = false;
  bool nez_calibrate = false;

  /* Parse switches */
  iarg=1;
  while (iarg < argc && *argv[iarg] == '-') {
    if (strcmp(argv[iarg], "-verbose") == 0) {
      verbose = true;
    }
    else if (strcmp(argv[iarg], "-onefield") == 0) {
      one_field = true;
    }
    else if (strcmp(argv[iarg], "-oneplot") == 0) {
      one_plot = true;
      one_field = true;
    }
    else if (strcmp(argv[iarg], "-png") == 0) {
      format = FORMAT_PNG;
    }
    else if ((strcmp(argv[iarg],"-ps") == 0) 
	     || (strcmp(argv[iarg],"-eps") == 0)) {
      format = FORMAT_POSTSCRIPT;
    }
    else if (strcmp(argv[iarg],"-hourrange") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&starthour) < 1) {
	cerr << "Failed to read start hour as float\n";
	exit(-1);
      }
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&endhour) < 1) {
	cerr << "Failed to read end hour as float\n";
	exit(-1);
      }
      use_duration = false;
    }
    else if (strcmp(argv[iarg],"-range") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&minval_override) < 1) {
	cerr << "Failed to read colour scale range\n";
	exit(-1);
      }
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&maxval_override) < 1) {
	cerr << "Failed to read colour scale range\n";
	exit(-1);
      }
      override_range = true;
    }
    else if (strcmp(argv[iarg],"-Zrange") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&minZ) < 1) {
	cerr << "Failed to read minimum Z\n";
	exit(-1);
      }
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&maxZ) < 1) {
	cerr << "Failed to read maximum Z\n";
	exit(-1);
      }
    }
    else if (strcmp(argv[iarg],"-vrange") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&minv) < 1) {
	cerr << "Failed to read minimum velocity\n";
	exit(-1);
      }
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&maxv) < 1) {
	cerr << "Failed to read maximum velocity\n";
	exit(-1);
      }
    }
    else if (strcmp(argv[iarg],"-widthrange") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&minwidth) < 1) {
	cerr << "Failed to read minimum spectral width\n";
	exit(-1);
      }
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&maxwidth) < 1) {
	cerr << "Failed to read maximum spectral width\n";
	exit(-1);
      }
    }
    else if (strcmp(argv[iarg],"-svrange") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&minsv) < 1) {
	cerr << "Failed to read minimum velocity standard deviation\n";
	exit(-1);
      }
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&maxsv) < 1) {
	cerr << "Failed to read maximum velocity standard deviation\n";
	exit(-1);
      }
    }
    else if (strcmp(argv[iarg],"-heightrange") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&minheight) < 1) {
	cerr << "Failed to read minimum height as float\n";
	exit(-1);
      }
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&maxheight) < 1) {
	cerr << "Failed to read maximum height as float";
	exit(-1);
      }
    }
    else if (strcmp(argv[iarg],"-duration") == 0) {
      if (argc > iarg+1) {
	duration = atof(argv[++iarg]);
	if (duration == 0.0) {
	  use_duration = false;
	}
	else {
	  use_duration = true;
	}
      }
      else {
	cerr << "Duration not specified\n";
	exit(-1);
      }
    }
    else if (strcmp(argv[iarg],"-nezcal") == 0) {
      if (argc > iarg+1) {
	nez_override = atof(argv[++iarg]);
	nez_calibrate = true;
      }
      else {
	cerr << "Noise-equivalent reflectivity for calibration not specified\n";
	exit(-1);
      }
    }
    else if (strcmp(argv[iarg],"-betacal") == 0) {
      if (argc > iarg+1) {
	betacal = 0.1 * atof(argv[++iarg]);
      }
      else {
	cerr << "Beta calibration figure not specified\n";
	exit(-1);
      }
    }
    else if (strcmp(argv[iarg],"-rayduration") == 0) {
      if (argc > iarg+1) {
	rayduration = atof(argv[++iarg])/3600.0;
	if (rayduration <= 0.0) {
	  cerr << "Warning: ray duration less than or equal to 0; using "
	       << DEFAULT_RAY_DURATION*3600.0 << " seconds";
	  rayduration = DEFAULT_RAY_DURATION;
	}
      }
      else {
	cerr << "Ray duration not specified\n";
	exit(-1);
      }
    }
    else if (strcmp(argv[iarg], "-size") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%dx%d",
				   &width,&height) < 2) {
	cerr << "Failed to read size: should be in the form "
	  "<width>x<height>\n";
	exit(-1);
      }
    }
    else if (strcmp(argv[iarg], "-barsize") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%dx%d",
				   &bar_width,&bar_height) < 2) {
	cerr << "Failed to read bar size: should be in the form "
	  "<width>x<height>\n";
	exit(-1);
      }
      if (bar_width < 1 || bar_height < 1) {
	no_bar = true;
      }
    }
    else if (strcmp(argv[iarg], "-fontsize") == 0) {
      if (argc <= iarg+1) {
	cerr << "Failed to read font size\n";
	exit(-1);
      }
      font_size = atoi(argv[++iarg]);
      if (font_size < 1) {
	cerr << "Font size must be a positive integer\n";
	exit(-1);
      }
    }
    else if (strcmp(argv[iarg], "-units") == 0) {
      if (argc <= iarg+1) {
	cerr << "Failed to read units\n";
	exit(-1);
      }
      units_override = argv[++iarg];
    }
    else if (strcmp(argv[iarg], "-system") == 0) {
      if (argc <= iarg+1) {
	cerr << "Failed to read system name\n";
	exit(-1);
      }
      system_override = argv[++iarg];
    }
    else if (strcmp(argv[iarg], "-field") == 0) {
      if (argc <= iarg+1) {
	cerr << "Failed to read field name\n";
	exit(-1);
      }
      long_name_override = argv[++iarg];
    }
    else if (strcmp(argv[iarg], "-id") == 0) {
      if (argc <= iarg+1) {
	cerr << "Failed to read system ID\n";
	exit(-1);
      }
      system_code = argv[++iarg];
      underscore1 = "_";
    }
    else if (strcmp(argv[iarg],"-prefix") == 0) {
      if (argc > iarg+1) {
	prefix = argv[++iarg];
      }
      else {
	cerr << "Prefix not specified\n";
	exit(-1);
      }
    }
    else if (strcmp(argv[iarg],"-palette") == 0) {
      if (argc > iarg+1) {
	mapfile = argv[++iarg];
      }
      else {
	cerr << "Map file not specified\n";
	exit(-1);
      }
    }
    else if (strcmp(argv[iarg],"-xfs") == 0) {
      if (argc > iarg+1) {
	xfs = argv[++iarg];
      }
      else {
	cerr << "X font server not specified";
	exit(-1);
      }
    }
    else {
      cerr << argv[iarg] << " not understood\n";
      Usage(argv);
    }
    ++iarg;
  }
  if (iarg > argc-1) {
    Usage(argv);
  }

  if (format == FORMAT_POSTSCRIPT) {
    suffix = "eps";
  }
  else {
    suffix = "png";
  }    

  if (no_bar) {
    fig_width = width + x_offset + 20;
  }
  else {
    fig_width = width + x_offset + bar_width + 64;
  }
  fig_height = height + y_offset + 30;

  try {

    /* Contact font server */
    font_loader = new RFontLoader(xfs, format == FORMAT_PNG);
    font_loader->verbose(verbose);

    if (mapfile) {
      if (colormap.load(mapfile)) {
	if (!quiet) {
	  cerr << "Warning: " << mapfile << " not found\n";
	}
	create_jet_palette(&colormap,64);
      }
    }
    else {
      create_jet_palette(&colormap,64);
    }

    /* Loop through all file names on the command line */
    for (ifile = iarg; ifile < argc; ++ifile) {
      /* Get dimension sizes */
      handle_error(nc_open(argv[ifile], NC_NOWRITE, &ncid));
      handle_error(nc_inq_dimid(ncid, "time", &timedimid));
      handle_error(nc_inq_dimlen(ncid, timedimid, &nrays));
      handle_error(nc_inq_dimid(ncid, "range", &rangedimid));
      handle_error(nc_inq_dimlen(ncid, rangedimid, &ngates));

      /* Get the ranges */
      handle_error(nc_inq_varid(ncid, "range", &rangeid));
      handle_error(nc_inq_varndims(ncid, rangeid, &tmp));
      if (tmp != 1) {
	cerr << "Range variable is not a 1D vector\n";
	exit(-1);
      } 
      range = new float[ngates];
      start[0] = 0;
      count[0] = ngates;
      handle_error(nc_get_vara_float(ncid, rangeid, start, count, range));

      /* Find time variable */
      handle_error(nc_inq_varid(ncid, "time", &timeid));

      /* Get the system attribute to use as the title */
      if (system_override) {
	system = system_override;
      }
      else {
	system = get_char_attribute(ncid, NC_GLOBAL, "system");
      }

      /* Get the date */
      day = get_short_attribute(ncid, NC_GLOBAL, "day");
      month = get_short_attribute(ncid, NC_GLOBAL, "month");
      year = get_short_attribute(ncid, NC_GLOBAL, "year");

      /* Set the hour range if necessary */
      if (use_duration) {
	start[0] = 0;
	handle_error(nc_get_var1_float(ncid, timeid, start, &starthour));
	if (starthour > 0.0 && starthour < duration*0.1) {
	  starthour = 0.0;
	}
	if (duration > 0.0) {
	  endhour = starthour + duration;
	}
	else {
	  start[0] = nrays-1;
	  handle_error(nc_get_var1_float(ncid, timeid, start, &endhour));
	  if (endhour < 24.0 && endhour-24.0 > duration*0.1) {
	    endhour = 24.0;
	  }
	  if (endhour-starthour < (-duration)) {
	    endhour = starthour - duration;
	  }
	  else if ((starthour = endhour + duration) < 0) {
	      starthour = 0;
	      endhour = -duration;
	  }	   
	}
      }

      /* How many variables do we need to look at? */
      handle_error(nc_inq_nvars(ncid, &nvars));
      max_vars = 0;

      /* Find 2D fields to plot */
      for (ivar = 0; ivar < nvars; ++ivar) {
	char *field;
	char name[NC_MAX_NAME];
	nc_type xtype;
	int ndims, natts;
	int dimids[NC_MAX_VAR_DIMS];

	nc_inq_var(ncid, ivar, name, &xtype,
		   &ndims, dimids, &natts);
	/* Is it 2D etc? */
	if (ndims == 2 && dimids[0] == timedimid && dimids[1] == rangedimid) {
	  RAxes *cbar;
	  int i = max_vars;
	  char tmpstr[256];
	  float scale_offset;

	  if (units_override) {
	    units = units_override;
	  }
	  else {
	    units = get_char_attribute(ncid, ivar, "units");
	  }

	  if (long_name_override) {
	    long_name = long_name_override;
	  }
	  else {
	    long_name = get_char_attribute(ncid, ivar, "long_name");
	  }

	  if ((missing_value[i] = get_float_attribute(ncid, ivar, "missing_value"))
	      < NO_FLOAT_ATTRIBUTE_TEST) {
	    cerr << "Warning: missing_value attribute not found for " << long_name << "\n";
	  }

	  if (one_field) {
	    field = "";
	    underscore2 = "";
	  }
	  else {
	    field = name;
	    underscore2 = "_";
	  }

	  if (!one_plot) {
	    sprintf(tmpstr, "%s/%04hd%02hd%02hd%s%s%s%s.%s",
		    prefix, year, month, day, underscore1,
		    system_code, underscore2, field, suffix);
	    ofs[i].open(tmpstr);
	    if (!ofs[i]) {
	      cerr << "Failed to write " << tmpstr << "\n";
	      exit(-1);
	    }
	    cerr << "Writing " << tmpstr << "\n";

	    /* Initialise plot */
	    if (format == FORMAT_POSTSCRIPT) {
	      figure[i] = new RFigure(fig_width, fig_height, POSTSCRIPT,
				      ofs+i, system, 0.5);
	    }
	    else {
	      figure[i] = new RFigure(fig_width, fig_height, BITMAP_8BIT);
	    }
	  }
	  else /* One plot to standard output */ {
	    /* Initialise plot */
	    if (format == FORMAT_POSTSCRIPT) {
	      figure[i] = new RFigure(fig_width, fig_height, POSTSCRIPT,
				      &cout, system, 0.5);
	    }
	    else {
	      figure[i] = new RFigure(fig_width, fig_height, BITMAP_8BIT);
	    }
	  }

	  figure[i]->clear();
	  figure[i]->axis(0,0,fig_width, fig_height);
	  figure[i]->font = font_loader->load(font_size,
					      string("helvetica"),
					      string("bold"),
					      string("r"),
					      string("adobe"));
	  figure[i]->text(0.01, 0.0, system, ANCHOR_BOTTOMLEFT);
	  figure[i]->font = font_loader->load(font_size,
					      string("helvetica"),
					      string("medium"),
					      string("r"),
					      string("adobe"));

	  axes[i] = new RAxes(figure[i], x_offset, y_offset, 
			      x_offset + width, y_offset + height);

	  axes[i]->axis(starthour, minheight, endhour, maxheight);
	  get_date(tmpstr, day, month, year);
	  axes[i]->text(starthour, maxheight, tmpstr, ANCHOR_BOTTOMLEFT);
	  axes[i]->clipbox_to_axis();
  
	  log_scale[i] = false;
	  using_nez[i] = false;
	  vfold[i] = NO_FLOAT_ATTRIBUTE;
	  minvalue[i] = minZ;
	  maxvalue[i] = maxZ;
	  interval[i] = 10;
	  scale_offset = 0;

	  if (strcmp(name, "beta") == 0) {
	    log_scale[i] = true;
	    minvalue[i] = 0;
	    maxvalue[i] = 3;
	    interval[i] = 1;
	    scale_offset = -7;
	  }
	  else if (strcmp(name, "v") == 0) {
	    minvalue[i] = minv;
	    maxvalue[i] = maxv;
	    interval[i] = 1;
	    vfold[i] = get_float_attribute(ncid, ivar, "folding_velocity");
	  }
	  else if (strcmp(name, "sigma_v") == 0) {
	    /*
	    minvalue[i] = minsv;
	    maxvalue[i] = maxsv;
	    if (maxsv - minsv <= 1.1) {
	      interval[i] = 0.2;
	    }
	    else {
	      interval[i] = 0.5;
	    }
	    */
	    log_scale[i] = true;
	    minvalue[i] = minsv;
	    maxvalue[i] = maxsv;
	    interval[i] = 1;
	  }
	  else if (strcmp(name, "width_h") == 0) {
	    /*
	    minvalue[i] = minwidth;
	    maxvalue[i] = maxwidth;
	    if (maxwidth - minwidth <= 1.1) {
	      interval[i] = 0.2;
	    }
	    else {
	      interval[i] = 0.5;
	    }
	    */
	    log_scale[i] = true;
	    minvalue[i] = minwidth;
	    maxvalue[i] = maxwidth;
	    interval[i] = 0.5;
	  }
	  else if (nez_calibrate && strcmp(name, "Zh") == 0) {
	    if (nc_inq_varid(ncid, "nez", &nezid) == NC_NOERR) {
	      using_nez[i] = true;
	    }	    
	  }

	  if (override_range) {
	    minvalue[i] = minval_override - scale_offset;
	    maxvalue[i] = maxval_override - scale_offset;
	  }

	  if (!no_bar) {
	    cbar = new RAxes(figure[i], (fig_width-10)-bar_width,
			     y_offset+0.5*(height-bar_height),
			     fig_width-10, y_offset+0.5*(height+bar_height));
	    cbar->colorbar(&colormap,VERTICAL);
	    cbar->box();
	    cbar->axis(0,minvalue[i]+scale_offset,1,maxvalue[i]+scale_offset);
	    if (log_scale[i]) {
	      cbar->yticks(minvalue[i]+scale_offset,interval[i],"10<SUP>%g",0);
	    }
	    else {
	      cbar->yticks(minvalue[i]+scale_offset,interval[i],"%g",0);
	    }
	    sprintf(tmpstr,"%s (%s)", long_name, units);
	    cbar->ylabel(tmpstr);
	    delete cbar;
	  }

	  varid[i] = ivar;
	  ++max_vars;
	  if (max_vars >= MAX_PLOTS) {
	    break;
	  }
	}
	if (max_vars == 1 && one_field) {
	  break;
	}
      }

      var = new float[ngates];
      start[1] = 0;
      count[0] = 1;
      count[1] = ngates;

      for (iray = 0; iray < nrays; ++iray) {
	start[0] = iray;
	handle_error(nc_get_var1_float(ncid, timeid, start, &time));
	if (iray < nrays-1) {
	  start[0] = iray+1;
	  handle_error(nc_get_var1_float(ncid, timeid, start, &nexttime));
	  if (nexttime - time > rayduration || nexttime < time) {
	    nexttime = time + rayduration;
	  }
	  start[0] = iray;
	}
	else {
	  nexttime = time + rayduration;
	}

	for (ivar = 0; ivar < max_vars; ++ivar) {
	  float gap = missing_value[ivar];
	  handle_error(nc_get_vara_float(ncid, varid[ivar],
					 start, count, var));
	  if (using_nez[ivar]) {
	    handle_error(nc_get_var1_float(ncid, nezid, start, &nez));
	    for (unsigned int igate = 0; igate < ngates; ++igate) {
	      if (var[igate] != gap) {
		var[igate] += (nez_override - nez);
	      }
	    }	    
	  }
	  else if (log_scale[ivar]) {
	    for (unsigned int igate = 0; igate < ngates; ++igate) {
	      gap = -1.0;
	      if (var[igate] <= 0) {
		var[igate] = gap;
	      }
	      else {
		var[igate] = log10(var[igate]) + betacal;
	      }
	    }
	  }
	  else if (vfold[ivar] > NO_FLOAT_ATTRIBUTE_TEST) {
	    for (unsigned int igate = 0; igate < ngates; ++igate) {
	      float tmpvar = var[igate];
	      if (tmpvar != gap && tmpvar < minvalue[ivar]
		  && tmpvar < maxvalue[ivar]) {
		var[igate] += 2 * vfold[ivar];
	      }
	      else if (tmpvar != gap && tmpvar > maxvalue[ivar]
		       && tmpvar > minvalue[ivar]) {
		var[igate] -= 2 * vfold[ivar];
	      }
	    }
	  }
	  axes[ivar]->vray(time, nexttime,
			   range, var, ngates, &colormap, 
			   minvalue[ivar], maxvalue[ivar], gap);
	}
      }

      handle_error(nc_close(ncid));
      delete[] var;
      delete[] range; 
      if (system && !system_override) {
	delete[] system;
      }
      if (units && !units_override) {
	delete[] units;
      }
      if (long_name && !long_name_override) {
	delete[] long_name;
      }

      dhour = (endhour-starthour > 6) ? floor((endhour-starthour)
					      / 6.0) :
	(1.0+floor((endhour-starthour)/4.0))/2.0;
      for (ivar = 0; ivar < max_vars; ++ivar) {
	axes[ivar]->box();
	axes[ivar]->xticks_time(0, dhour);
	axes[ivar]->yticks(0,1);
	axes[ivar]->xlabel("Time (UTC)");
	axes[ivar]->ylabel("Height (km)");
	if (format == FORMAT_PNG) {
	  if (!one_plot) {
	    figure[ivar]->write_png(&ofs[ivar]);
	  }
	  else {
	    figure[ivar]->write_png(&cout);
	  }
	}
	delete axes[ivar];
	delete figure[ivar];
	if (!one_plot) {
	  ofs[ivar].close();
	}
      }
      if (one_plot) {
	break;
      }
    }
    delete font_loader;
  }
  catch (const char *error_message) {
    cerr << error_message << endl;
    exit(1);
  }
}
